/*
 * zame-js-z80
 * Copyright 2010, Slava Tretyak
 */

#define FLAG_C  0x01  // carry
#define FLAG_N  0x02  // subtract
#define FLAG_PV 0x04  // parity or overflow
#define FLAG_3  0x08  // copy of bit 3
#define FLAG_H  0x10  // half carry
#define FLAG_5  0x20  // copy of bit 5
#define FLAG_Z  0x40  // zero
#define FLAG_S  0x80  // sign

#define FLAG_C_M16 0x10000
#define FLAG_C_S16 0x10
#define FLAG_C_M8 0x100
#define FLAG_C_S8 8

#define FLAG_H_M16 0x1000
#define FLAG_H_S16 8

#define FLAG_PV_M8 0x80
#define FLAG_PV_S8 5
#define FLAG_PV_IFF_S8 2

#define FLAG_C_TO_H 4
#define FLAG_H_TO_C 4
#define FLAG_N_TO_5 4
#define FLAG_S_TO_N 6

#define CPU_C    0
#define CPU_B    1
#define CPU_E    2
#define CPU_D    3
#define CPU_L    4
#define CPU_H    5
#define CPU_F    6
#define CPU_A    7
#define CPU_IXL  8
#define CPU_IXH  9
#define CPU_IYL  10
#define CPU_IYH  11
#define CPU_SPL  12
#define CPU_SPH  13
#define CPU_PCL  14
#define CPU_PCH  15
#define CPU_MPL  16
#define CPU_MPH  17
#define CPU_C_   18
#define CPU_B_   19
#define CPU_E_   20
#define CPU_D_   21
#define CPU_L_   22
#define CPU_H_   23
#define CPU_F_   24
#define CPU_A_   25
#define CPU_I    26
#define CPU_R    27
#define CPU_IFF1 28
#define CPU_IFF2 29
#define CPU_IM   30

#define REG_A      this.regs[CPU_A]
#define REG_F      this.regs[CPU_F]
#define REG_B      this.regs[CPU_B]
#define REG_C      this.regs[CPU_C]
#define REG_D      this.regs[CPU_D]
#define REG_E      this.regs[CPU_E]
#define REG_H      this.regs[CPU_H]
#define REG_L      this.regs[CPU_L]
#define REG_IXH    this.regs[CPU_IXH]
#define REG_IXL    this.regs[CPU_IXL]
#define REG_IYH    this.regs[CPU_IYH]
#define REG_IYL    this.regs[CPU_IYL]
#define REG_SPH    this.regs[CPU_SPH]
#define REG_SPL    this.regs[CPU_SPL]
#define REG_PCH    this.regs[CPU_PCH]
#define REG_PCL    this.regs[CPU_PCL]
#define REG_MPH    this.regs[CPU_MPH]
#define REG_MPL    this.regs[CPU_MPL]
#define REG_A_     this.regs[CPU_A_]
#define REG_F_     this.regs[CPU_F_]
#define REG_B_     this.regs[CPU_B_]
#define REG_C_     this.regs[CPU_C_]
#define REG_D_     this.regs[CPU_D_]
#define REG_E_     this.regs[CPU_E_]
#define REG_H_     this.regs[CPU_H_]
#define REG_L_     this.regs[CPU_L_]
#define REG_I      this.regs[CPU_I]
#define REG_R      this.regs[CPU_R]
#define REG_IFF1   this.regs[CPU_IFF1]
#define REG_IFF2   this.regs[CPU_IFF2]
#define REG_IM     this.regs[CPU_IM]

#define GET_AF     (this.regs[CPU_A] << 8 | this.regs[CPU_F])
#define GET_BC     (this.regs[CPU_B] << 8 | this.regs[CPU_C])
#define GET_DE     (this.regs[CPU_D] << 8 | this.regs[CPU_E])
#define GET_HL     (this.regs[CPU_H] << 8 | this.regs[CPU_L])
#define GET_IX     (this.regs[CPU_IXH] << 8 | this.regs[CPU_IXL])
#define GET_IY     (this.regs[CPU_IYH] << 8 | this.regs[CPU_IYL])
#define GET_SP     (this.regs[CPU_SPH] << 8 | this.regs[CPU_SPL])
#define GET_PC     (this.regs[CPU_PCH] << 8 | this.regs[CPU_PCL])
#define GET_MP     (this.regs[CPU_MPH] << 8 | this.regs[CPU_MPL])
#define GET_AF_    (this.regs[CPU_A_] << 8 | this.regs[CPU_F_])
#define GET_BC_    (this.regs[CPU_B_] << 8 | this.regs[CPU_C_])
#define GET_DE_    (this.regs[CPU_D_] << 8 | this.regs[CPU_E_])
#define GET_HL_    (this.regs[CPU_H_] << 8 | this.regs[CPU_L_])

function CONV_INT8(v) { return v > 127 ? v - 256 : v; }
function CONV_INT16(v) { return v > 32767 ? v - 65536 : v; }

#define CPU_DO_READ_WORD \
	this.tmp_word = this.read_byte(); \
	this.tmp_word |= this.read_byte() << 8;

// priority '+' -> '&'
#define CPU_DO_READ_OFFSET(GET_RP) \
	this.tmp_int8 = this.read_offset(); \
	this.SET_MP(GET_RP + this.tmp_int8 & 0xFFFF);

// priority '+' -> '&' -> '|'
#define CPU_INC_R \
	REG_R = REG_R & 0x80 | REG_R + 1 & 0x7F;

#include "op_common.js"
#include "op_pref_00.js"
#include "op_pref_CB.js"
#include "op_pref_DD.js"
#include "op_pref_DD_CB.js"
#include "op_pref_ED.js"
#include "op_pref_FD.js"
#include "op_pref_FD_CB.js"

function Cpu(ptr_read, ptr_read_m1, ptr_write, ptr_in, ptr_out, ptr_read_int)
{
	this.regs = [
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	]

	this.SET_AF = function(v)  { this.regs[CPU_A] = v >> 8; this.regs[CPU_F] = v & 0xFF; }
	this.SET_BC = function(v)  { this.regs[CPU_B] = v >> 8; this.regs[CPU_C] = v & 0xFF; }
	this.SET_DE = function(v)  { this.regs[CPU_D] = v >> 8; this.regs[CPU_E] = v & 0xFF; }
	this.SET_HL = function(v)  { this.regs[CPU_H] = v >> 8; this.regs[CPU_L] = v & 0xFF; }
	this.SET_IX = function(v)  { this.regs[CPU_IXH] = v >> 8; this.regs[CPU_IXL] = v & 0xFF; }
	this.SET_IY = function(v)  { this.regs[CPU_IYH] = v >> 8; this.regs[CPU_IYL] = v & 0xFF; }
	this.SET_SP = function(v)  { this.regs[CPU_SPH] = v >> 8; this.regs[CPU_SPL] = v & 0xFF; }
	this.SET_PC = function(v)  { this.regs[CPU_PCH] = v >> 8; this.regs[CPU_PCL] = v & 0xFF; }
	this.SET_MP = function(v)  { this.regs[CPU_MPH] = v >> 8; this.regs[CPU_MPL] = v & 0xFF; }
	this.SET_AF_ = function(v) { this.regs[CPU_A_] = v >> 8; this.regs[CPU_F_] = v & 0xFF; }
	this.SET_BC_ = function(v) { this.regs[CPU_B_] = v >> 8; this.regs[CPU_C_] = v & 0xFF; }
	this.SET_DE_ = function(v) { this.regs[CPU_D_] = v >> 8; this.regs[CPU_E_] = v & 0xFF; }
	this.SET_HL_ = function(v) { this.regs[CPU_H_] = v >> 8; this.regs[CPU_L_] = v & 0xFF; }

	this.is_halted = false;
	this.is_opcode = false;
	this.is_noint = false;
	this.is_reset_pv = false;
	this.is_read_int = false

	this.ptr_read = ptr_read;			// (word addr)
	this.ptr_read_m1 = ptr_read_m1;		// (word addr)
	this.ptr_write = ptr_write;			// (word addr, byte val)
	this.ptr_in = ptr_in;				// (word port)
	this.ptr_out = ptr_out;				// (word port, byte val)
	this.ptr_read_int = ptr_read_int;	// ()

	this.tick = null;
	this.optable = null;
	this.prefix = 0;
	this.tstate = 0;

	this.tmp_int8 = 0;
	this.tmp_int16 = 0;
	this.tmp_int32 = 0;
	this.tmp_byte = 0;
	this.tmp_byte_b = 0;
	this.tmp_addr = 0;
	this.tmp_word = 0;
	this.tmp_word_b = 0;
	this.tmp_dword = 0;

	this.get_rp = function(idx)
	{
		return this.regs[idx + 1] << 8 | this.regs[idx];
	}

	this.set_rp = function(idx, v)
	{
		this.regs[idx + 1] = v >> 8;
		this.regs[idx] = v & 0xFF;
	}

	this.read_byte = function()
	{
		if (this.is_read_int) return this.ptr_read_int();

		var ret = this.ptr_read(GET_PC);
		this.SET_PC(GET_PC + 1 & 0xFFFF);

		return ret;
	}

	this.read_offset = function()
	{
		var off = this.read_byte();
		return off > 127 ? off - 256 : off;
	}

	this.reset = function()
	{
		for (var i = 0; i < this.regs.length; i++) this.regs[i] = 0;

		this.SET_AF(0xFFFF);
		this.SET_SP(0xFFFF);

		// really all others registers may has any value
		this.SET_BC(0xFFFF);
		this.SET_DE(0xFFFF);
		this.SET_HL(0xFFFF);
		this.SET_IX(0xFFFF);
		this.SET_IY(0xFFFF);
		this.SET_AF_(0xFFFF);
		this.SET_BC_(0xFFFF);
		this.SET_DE_(0xFFFF);
		this.SET_HL_(0xFFFF);

		this.is_halted = false;
		this.is_opcode = false;
		this.is_noint = false;
		this.is_reset_pv = false;
		this.is_read_int = false;

		this.tick = Cpu.tick_def;
		this.optable = optable_00;
		this.prefix = 0;
		this.tstate = 0;
	}

	this.do_int = function()
	{
		if (!REG_IFF1 || this.is_noint || this.is_opcode || this.prefix) return 0;

		if (this.is_halted)
		{
			this.SET_PC(GET_PC + 1 & 0xFFFF);
			this.is_halted = false;
		}

		REG_IFF1 = 0;
		REG_IFF2 = 0;

		if (this.is_reset_pv)
		{
			REG_F &= ~FLAG_PV;
			this.is_reset_pv = false;
		}

		this.is_opcode = true;
		this.tstate = 0;

		switch (REG_IM)
		{
			case 0:
			{
				this.is_read_int = true;
				Cpu.tick_int.call(this);	// tstate changed in this function

				if (this.prefix) this.tick = Cpu.tick_int;
			}
			break;

			case 1:
			{
				CPU_INC_R;
				this.tstate += 6;				// (4 + 2)
				optable_00[0xFF].call(this);	// RST #38
			}
			break;

			case 2:
			{
				CPU_INC_R;
				var vec = this.ptr_read_int();
				var addr = REG_I << 8 | vec;

				this.SET_SP(GET_SP - 1 & 0xFFFF); this.ptr_write(GET_SP, REG_PCH);
				this.SET_SP(GET_SP - 1 & 0xFFFF); this.ptr_write(GET_SP, REG_PCL);

				REG_PCL = this.ptr_read(addr);
				REG_PCH = this.ptr_read(addr + 1 & 0xFFFF);

				this.SET_MP(GET_PC);
				this.tstate += 19;
			}
			break;
		}

		this.is_opcode = false;
		return this.tstate;
	}

	this.do_nmi = function()
	{
		if (this.is_noint || this.is_opcode || this.prefix) return 0;

		if (this.is_halted)
		{
			this.SET_PC(GET_PC + 1 & 0xFFFF);
			this.is_halted = false;
		}

		this.is_opcode = true;
		CPU_INC_R;
		REG_IFF1 = 0;

		this.SET_SP(GET_SP - 1 & 0xFFFF); this.ptr_write(GET_SP, REG_PCH);
		this.SET_SP(GET_SP - 1 & 0xFFFF); this.ptr_write(GET_SP, REG_PCL);

		this.SET_PC(0x0066);
		this.SET_MP(0x0066);		// SET_MP(GET_PC)
		this.is_opcode = false;

		return 11;
	}

	this.is_int_possible = function()
	{
		return (REG_IFF1 && !this.is_noint && !this.is_opcode && !this.prefix);
	}

	this.is_nmi_possible = function()
	{
		return (!this.is_noint && !this.is_opcode && !this.prefix);
	}

	Cpu.init_tables();
	this.reset();
}

Cpu.is_tbl_initialized = false;
Cpu.tbl_parity = [];	// [0x100]

Cpu.init_tables = function()
{
	if (Cpu.is_tbl_initialized) return;

	for (var i = 0; i < 0x100; i++)
	{
		var p = ((i & 0x80) >> 7) ^ ((i & 0x40) >> 6) ^ ((i & 0x20) >> 5) ^ ((i & 0x10) >> 4) ^ ((i & 0x08) >> 3) ^ ((i & 0x04) >> 2) ^ ((i & 0x02) >> 1) ^ (i & 0x01);
		Cpu.tbl_parity.push(p ? 0 : FLAG_PV);
	}

	Cpu.is_tbl_initialized = true;
}

Cpu.tick_def = function()
{
	this.is_opcode = true;
	this.is_noint = false;
	this.is_reset_pv = false;
	this.tstate = 4;

	var op = this.ptr_read_m1(GET_PC);
	this.SET_PC(GET_PC + 1 & 0xFFFF);
	CPU_INC_R;

	this.optable[op].call(this);
	this.is_opcode = false;

	return this.tstate;
}

Cpu.tick_int = function()
{
	this.is_opcode = true;
	this.is_noint = false;
	this.is_reset_pv = false;
	this.tstate = 6;	// (4 + 2)

	var op = this.ptr_read_int();
	CPU_INC_R;

	this.optable[op].call(this);
	this.is_opcode = false;

	if (this.prefix) {
		this.is_noint = true;
	} else {
		this.tick = Cpu.tick_def;
		this.is_read_int = false;
	}

	return this.tstate;
}

Cpu.reg_bc = CPU_C;
Cpu.reg_c = CPU_C;
Cpu.reg_b = CPU_B;
Cpu.reg_de = CPU_E;
Cpu.reg_e = CPU_E;
Cpu.reg_d = CPU_D;
Cpu.reg_hl = CPU_L;
Cpu.reg_l = CPU_L;
Cpu.reg_h = CPU_H;
Cpu.reg_af = CPU_F;
Cpu.reg_f = CPU_F;
Cpu.reg_a = CPU_A;
Cpu.reg_ix = CPU_IXL;
Cpu.reg_ixl = CPU_IXL;
Cpu.reg_ixh = CPU_IXH;
Cpu.reg_iy = CPU_IYL;
Cpu.reg_iyl = CPU_IYL;
Cpu.reg_iyh = CPU_IYH;
Cpu.reg_sp = CPU_SPL;
Cpu.reg_spl = CPU_SPL;
Cpu.reg_sph = CPU_SPH;
Cpu.reg_pc = CPU_PCL;
Cpu.reg_pcl = CPU_PCL;
Cpu.reg_pch = CPU_PCH;
Cpu.reg_mp = CPU_MPL;
Cpu.reg_mpl = CPU_MPL;
Cpu.reg_mph = CPU_MPH;
Cpu.reg_bc_ = CPU_C_;
Cpu.reg_c_ = CPU_C_;
Cpu.reg_b_ = CPU_B_;
Cpu.reg_de_ = CPU_E_;
Cpu.reg_e_ = CPU_E_;
Cpu.reg_d_ = CPU_D_;
Cpu.reg_hl_ = CPU_L_;
Cpu.reg_l_ = CPU_L_;
Cpu.reg_h_ = CPU_H_;
Cpu.reg_af_ = CPU_F_;
Cpu.reg_f_ = CPU_F_;
Cpu.reg_a_ = CPU_A_;
Cpu.reg_i = CPU_I;
Cpu.reg_r = CPU_R;
Cpu.reg_iff1 = CPU_IFF1;
Cpu.reg_iff2 = CPU_IFF2;
Cpu.reg_im = CPU_IM;
