/*
 * zame-js-z80
 * Copyright 2010, Slava Tretyak
 */

#define OP_NOP(FN,...) \
	function FN() { \
		__VA_ARGS__ \
	}

#define OP_LD_RP_NN(FN,SET_RP,...) \
	function FN() { \
		CPU_DO_READ_WORD; \
		SET_RP(this.tmp_word); \
		this.tstate += (10-4); \
		__VA_ARGS__ \
	}

#define OP_LD_MRP_A(FN,GET_RP,...) \
	function FN() { \
		REG_MPH = REG_A; \
		REG_MPL = GET_RP + 1 & 0xFF; \
		this.ptr_write(GET_RP, REG_A); \
		this.tstate += (7-4); \
		__VA_ARGS__ \
	}

#define OP_DO_RP(FN,DO,SET_RP,GET_RP,...) \
	function FN() { \
		DO(SET_RP, GET_RP); \
		this.tstate += (6-4); \
		__VA_ARGS__ \
	}

#define OP_LD_R_N(FN,R,...) \
	function FN() { \
		R = this.read_byte(); \
		this.tstate += (7-4); \
		__VA_ARGS__ \
	}

#define OP_ADD_RP_RP(FN,SET_RP1,GET_RP1,GET_RP2,...) \
	function FN() { \
		DO_ADD_16(SET_RP1, GET_RP1, GET_RP2); \
		this.tstate += (11-4); \
		__VA_ARGS__ \
	}

#define OP_LD_A_MRP(FN,GET_RP,...) \
	function FN() { \
		REG_A = this.ptr_read(GET_RP); \
		this.SET_MP(GET_RP + 1 & 0xFFFF); \
		this.tstate += (7-4); \
		__VA_ARGS__ \
	}

#define OP_DO_R(FN,DO,R,...) \
	function FN() { \
		DO(R); \
		__VA_ARGS__ \
	}

#define OP_DO_MHL(FN,DO,...) \
	function FN() { \
		this.tmp_byte = this.ptr_read(GET_HL); \
		DO(this.tmp_byte); \
		this.ptr_write(GET_HL, this.tmp_byte); \
		this.tstate += (11-4); \
		__VA_ARGS__ \
	}

#define OP_DO_ORP(FN,DO,GET_RP) \
	function FN() { \
		CPU_DO_READ_OFFSET(GET_RP); \
		this.tmp_addr = GET_RP + this.tmp_int8 & 0xFFFF; \
		this.tmp_byte = this.ptr_read(this.tmp_addr); \
		DO(this.tmp_byte); \
		this.ptr_write(this.tmp_addr, this.tmp_byte); \
		this.tstate += (19-4); \
		DO_PREF_00; \
	}

#define OP_LD_DO_R_PORP(FN,DO,R,GET_RP) \
	function FN() { \
		this.tmp_addr = GET_RP + this.tmp_int8 & 0xFFFF; \
		this.SET_MP(this.tmp_addr); \
		this.tmp_byte = this.ptr_read(this.tmp_addr); \
		DO(this.tmp_byte); \
		R = this.tmp_byte; \
		this.ptr_write(this.tmp_addr, this.tmp_byte); \
		this.tstate += (19-4); \
		DO_PREF_00; \
	}

#define OP_DO_PORP(FN,DO,GET_RP) \
	function FN() { \
		this.tmp_addr = GET_RP + this.tmp_int8 & 0xFFFF; \
		this.SET_MP(this.tmp_addr); \
		this.tmp_byte = this.ptr_read(this.tmp_addr); \
		DO(this.tmp_byte); \
		this.ptr_write(this.tmp_addr, this.tmp_byte); \
		this.tstate += (19-4); \
		DO_PREF_00; \
	}

#define OP_RLCA(FN,...) \
	function FN() { \
		REG_A = REG_A << 1 & 0xFF | REG_A >> 7; \
		REG_F = REG_F & (FLAG_PV | FLAG_Z | FLAG_S) | REG_A & (FLAG_C | FLAG_3 | FLAG_5); \
		__VA_ARGS__ \
	}

#define OP_RRCA(FN,...) \
	function FN() { \
		REG_F = REG_F & (FLAG_PV | FLAG_Z | FLAG_S) | REG_A & FLAG_C; \
		REG_A = REG_A >> 1 | REG_A << 7 & 0xFF; \
		REG_F |= REG_A & (FLAG_3 | FLAG_5); \
		__VA_ARGS__ \
	}

#define OP_RLA(FN,...) \
	function FN() { \
		this.tmp_byte = REG_A; \
		REG_A = REG_A << 1 & 0xFF | REG_F & FLAG_C; \
		REG_F = REG_F & (FLAG_PV | FLAG_Z | FLAG_S) | REG_A & (FLAG_3 | FLAG_5) | this.tmp_byte >> 7; \
		__VA_ARGS__ \
	}

#define OP_RRA(FN,...) \
	function FN() { \
		this.tmp_byte = REG_A; \
		REG_A = REG_A >> 1 | REG_F << 7 & 0xFF; \
		REG_F = REG_F & (FLAG_PV | FLAG_Z | FLAG_S) | REG_A & (FLAG_3 | FLAG_5) | this.tmp_byte & FLAG_C; \
		__VA_ARGS__ \
	}

#define OP_EX_AF_AF_(FN,...) \
	function FN() { \
		this.tmp_word = GET_AF; \
		this.SET_AF(GET_AF_); \
		this.SET_AF_(this.tmp_word); \
		__VA_ARGS__ \
	}

#define OP_DJNZ(FN,...) \
	function FN() { \
		this.tmp_int8 = this.read_byte(); \
		if (this.tmp_int8 > 127) this.tmp_int8 -= 256; \
		REG_B = REG_B - 1 & 0xFF; \
		if (REG_B) { \
			this.SET_PC(GET_PC + this.tmp_int8 & 0xFFFF); \
			this.SET_MP(GET_PC); \
			this.tstate += (13-4); \
		} else { \
			this.tstate += (8-4); \
		} \
		__VA_ARGS__ \
	}

#define OP_JR(FN,...) \
	function FN() { \
		this.tmp_int8 = this.read_byte(); \
		if (this.tmp_int8 > 127) this.tmp_int8 -= 256; \
		this.SET_PC(GET_PC + this.tmp_int8  & 0xFFFF); \
		this.SET_MP(GET_PC); \
		this.tstate += (12-4); \
		__VA_ARGS__ \
	}

#define OP_LD_R_R(FN,R1,R2,...) \
	function FN() { \
		R1 = R2; \
		__VA_ARGS__ \
	}

#define OP_LD_R_MRP(FN,R,GET_RP) \
	function FN() { \
		R = this.ptr_read(GET_RP); \
		this.tstate += (7-4); \
	}

#define OP_LD_R_ORP(FN,R,GET_RP) \
	function FN() { \
		CPU_DO_READ_OFFSET(GET_RP); \
		R = this.ptr_read(GET_RP + this.tmp_int8 & 0xFFFF); \
		this.tstate += (15-4); \
		DO_PREF_00; \
	}

#define OP_LD_MRP_R(FN,GET_RP,R,...) \
	function FN() { \
		this.ptr_write(GET_RP, R); \
		this.tstate += (7-4); \
		__VA_ARGS__ \
	}

#define OP_LD_ORP_R(FN,GET_RP,R) \
	function FN() { \
		CPU_DO_READ_OFFSET(GET_RP); \
		this.ptr_write(GET_RP + this.tmp_int8 & 0xFFFF, R); \
		this.tstate += (15-4); \
		DO_PREF_00; \
	}

#define OP_HALT(FN,...) \
	function FN() { \
		this.is_halted = true; \
		this.SET_PC(GET_PC - 1 & 0xFFFF); \
		__VA_ARGS__ \
	}

#define OP_JR_CC(FN,CC,...) \
	function FN() { \
		this.tmp_int8 = this.read_byte(); \
		if (this.tmp_int8 > 127) this.tmp_int8 -= 256; \
		if (CC) { \
			this.SET_PC(GET_PC + this.tmp_int8 & 0xFFFF); \
			this.SET_MP(GET_PC); \
			this.tstate += (12-4); \
		} else {\
			this.tstate += (7-4); \
		} \
		__VA_ARGS__ \
	}

#define OP_LD_MNN_RP(FN,RL,RH,...) \
	function FN() { \
		CPU_DO_READ_WORD; \
		this.SET_MP(this.tmp_word + 1 & 0xFFFF); \
		this.ptr_write(this.tmp_word, RL); \
		this.ptr_write(this.tmp_word + 1 & 0xFFFF, RH); \
		this.tstate += (16-4); \
		__VA_ARGS__ \
	}

#define OP_DAA(FN,...) \
	function FN() { \
		this.tmp_byte = REG_A; \
		if (REG_F & FLAG_N) { \
			if ((REG_F & FLAG_H) || ((REG_A & 0x0F) > 9)) this.tmp_byte -= 6; \
			if ((REG_F & FLAG_C) || (REG_A > 0x99)) this.tmp_byte -= 0x60; \
		} else { \
			if ((REG_F & FLAG_H) || ((REG_A & 0x0F) > 9)) this.tmp_byte += 6; \
			if ((REG_F & FLAG_C) || (REG_A > 0x99)) this.tmp_byte += 0x60; \
		} \
		this.tmp_byte &= 0xFF; \
		REG_F = REG_F & (FLAG_N | FLAG_C) \
				| this.tmp_byte & (FLAG_S | FLAG_5 | FLAG_3) \
				| (this.tmp_byte ? 0 : FLAG_Z) \
				| Cpu.tbl_parity[this.tmp_byte] \
				| (this.tmp_byte ^ REG_A) & FLAG_H \
				| (REG_A > 0x99 ? FLAG_C : 0); \
		REG_A = this.tmp_byte; \
		__VA_ARGS__ \
	}

#define OP_LD_RP_MNN(FN,RL,RH,...) \
	function FN() { \
		CPU_DO_READ_WORD; \
		RL = this.ptr_read(this.tmp_word); \
		RH = this.ptr_read(this.tmp_word + 1 & 0xFFFF); \
		this.SET_MP(this.tmp_word + 1 & 0xFFFF); \
		this.tstate += (16-4); \
		__VA_ARGS__ \
	}

#define OP_CPL(FN,...) \
	function FN() { \
		REG_A ^= 0xFF; \
		REG_F = REG_F & (FLAG_C | FLAG_PV | FLAG_Z | FLAG_S) | REG_A & (FLAG_3 | FLAG_5) | (FLAG_N | FLAG_H); \
		__VA_ARGS__ \
	}

#define OP_LD_MNN_A(FN,...) \
	function FN() { \
		CPU_DO_READ_WORD; \
		REG_MPH = REG_A; \
		REG_MPL = this.tmp_word + 1 & 0xFF; \
		this.ptr_write(this.tmp_word, REG_A); \
		this.tstate += (13-4); \
		__VA_ARGS__ \
	}

#define OP_LD_MHL_N(FN,...) \
	function FN() { \
		this.tmp_byte = this.read_byte(); \
		this.ptr_write(GET_HL, this.tmp_byte); \
		this.tstate += (10-4); \
		__VA_ARGS__ \
	}

#define OP_LD_ORP_N(FN,GET_RP) \
	function FN() { \
		CPU_DO_READ_OFFSET(GET_RP); \
		this.tmp_byte = this.read_byte(); \
		this.ptr_write(GET_RP + this.tmp_int8 & 0xFFFF, this.tmp_byte); \
		this.tstate += (15-4); \
		DO_PREF_00; \
	}

#define OP_SCF(FN,...) \
	function FN() { \
		REG_F = REG_F & (FLAG_PV | FLAG_Z | FLAG_S) | REG_A & (FLAG_5 | FLAG_3) | FLAG_C; \
		__VA_ARGS__ \
	}

#define OP_LD_A_MNN(FN,...) \
	function FN() { \
		CPU_DO_READ_WORD; \
		REG_A = this.ptr_read(this.tmp_word); \
		this.SET_MP(this.tmp_word + 1 & 0xFFFF); \
		this.tstate += (13-4); \
		__VA_ARGS__ \
	}

#define OP_CCF(FN,...) \
	function FN() { \
		REG_F = REG_F & (FLAG_PV | FLAG_Z | FLAG_S) | (REG_F & FLAG_C) << FLAG_C_TO_H | (REG_F & FLAG_C) ^ FLAG_C | REG_A & (FLAG_3 | FLAG_5); \
		__VA_ARGS__ \
	}

#define OP_DO_A_R(FN,DO,R,...) \
	function FN() { \
		DO(REG_A, R); \
		__VA_ARGS__ \
	}

#define OP_DO_A_MHL(FN,DO,...) \
	function FN() { \
		this.tmp_byte = this.ptr_read(GET_HL); \
		DO(REG_A, this.tmp_byte); \
		this.tstate += (7-4); \
		__VA_ARGS__ \
	}

#define OP_DO_A_ORP(FN,DO,GET_RP) \
	function FN() { \
		CPU_DO_READ_OFFSET(GET_RP); \
		this.tmp_byte = this.ptr_read(GET_RP + this.tmp_int8 & 0xFFFF); \
		DO(REG_A, this.tmp_byte); \
		this.tstate += (15-4); \
		DO_PREF_00; \
	}

#define OP_RET_CC(FN,CC,...) \
	function FN() { \
		if (CC) { \
			DO_RET; \
			this.tstate += (11-4); \
		} else { \
			this.tstate += (5-4); \
		} \
		__VA_ARGS__ \
	}

#define OP_POP_RP(FN,SET_RP,...) \
	function FN() { \
		DO_POP_TMP; \
		SET_RP(this.tmp_word); \
		this.tstate += (10-4); \
		__VA_ARGS__ \
	}

#define OP_JP_CC(FN,CC,...) \
	function FN() { \
		CPU_DO_READ_WORD; \
		this.SET_MP(this.tmp_word); \
		if (CC) { \
			this.SET_PC(this.tmp_word); \
		} \
		this.tstate += (10-4); \
		__VA_ARGS__ \
	}

#define OP_JP(FN,...) \
	function FN() { \
		CPU_DO_READ_WORD; \
		this.SET_PC(this.tmp_word); \
		this.SET_MP(this.tmp_word); \
		this.tstate += (10-4); \
		__VA_ARGS__ \
	}

#define OP_CALL_CC(FN,CC,...) \
	function FN() { \
		CPU_DO_READ_WORD; \
		this.SET_MP(this.tmp_word); \
		if (CC) { \
			DO_PUSH(GET_PC); \
			this.SET_PC(this.tmp_word); \
			this.tstate += (17-4); \
		} else { \
			this.tstate += (10-4); \
		} \
		__VA_ARGS__ \
	}

#define OP_PUSH_RP(FN,GET_RP,...) \
	function FN() { \
		DO_PUSH(GET_RP); \
		this.tstate += (11-4); \
		__VA_ARGS__ \
	}

#define OP_DO_A_N(FN,DO,...) \
	function FN() { \
		this.tmp_byte = this.read_byte(); \
		DO(REG_A, this.tmp_byte); \
		this.tstate += (7-4); \
		__VA_ARGS__ \
	}

#define OP_RST(FN,ADDR,...) \
	function FN() { \
		DO_PUSH(GET_PC); \
		this.SET_PC(ADDR); \
		this.SET_MP(ADDR); \
		this.tstate += (11-4); \
		__VA_ARGS__ \
	}

#define OP_RET(FN,...) \
	function FN() { \
		DO_RET; \
		this.tstate += (10-4); \
		__VA_ARGS__ \
	}

#define OP_CALL(FN,...) \
	function FN() { \
		CPU_DO_READ_WORD; \
		DO_PUSH(GET_PC); \
		this.SET_PC(this.tmp_word); \
		this.SET_MP(this.tmp_word); \
		this.tstate += (17-4); \
		__VA_ARGS__ \
	}

#define OP_OUT_N_A(FN,...) \
	function FN() { \
		this.tmp_word = REG_A << 8 | this.read_byte(); \
		DO_OUT(this.tmp_word, REG_A); \
		this.tstate += (11-4); \
		__VA_ARGS__ \
	}

#define OP_EXX(FN,...) \
	function FN() { \
		this.tmp_word = GET_BC; this.SET_BC(GET_BC_); this.SET_BC_(this.tmp_word); \
		this.tmp_word = GET_DE; this.SET_DE(GET_DE_); this.SET_DE_(this.tmp_word); \
		this.tmp_word = GET_HL; this.SET_HL(GET_HL_); this.SET_HL_(this.tmp_word); \
		__VA_ARGS__ \
	}

#define OP_IN_A_N(FN,...) \
	function FN() { \
		this.tmp_word = REG_A << 8 | this.read_byte(); \
		DO_IN(REG_A, this.tmp_word); \
		this.tstate += (11-4); \
		__VA_ARGS__ \
	}

#define OP_EX_SP_RP(FN,RL,RH,SET_RP,...) \
	function FN() { \
		this.tmp_word = this.ptr_read(GET_SP); \
		this.tmp_word |= this.ptr_read(GET_SP + 1 & 0xFFFF) << 8; \
		this.SET_MP(this.tmp_word); \
		this.ptr_write(GET_SP, RL); \
		this.ptr_write(GET_SP + 1 & 0xFFFF, RH); \
		SET_RP(this.tmp_word); \
		this.tstate += (19-4); \
		__VA_ARGS__ \
	}

#define OP_JP_RP(FN,GET_RP,...) \
	function FN() { \
		this.SET_PC(GET_RP); \
		__VA_ARGS__ \
	}

#define OP_EX_DE_HL(FN,...) \
	function FN() { \
		this.tmp_word = GET_DE; \
		this.SET_DE(GET_HL); \
		this.SET_HL(this.tmp_word); \
		__VA_ARGS__ \
	}

#define OP_DI(FN,...) \
	function FN() { \
		REG_IFF1 = 0; \
		REG_IFF2 = 0; \
		__VA_ARGS__ \
	}

#define OP_LD_SP_RP(FN,GET_RP,...) \
	function FN() { \
		this.SET_SP(GET_RP); \
		this.tstate += (6-4); \
		__VA_ARGS__ \
	}

#define OP_EI(FN,...) \
	function FN() { \
		REG_IFF1 = 1; \
		REG_IFF2 = 1; \
		this.is_noint = true; \
		__VA_ARGS__ \
	}

#define OP_BIT_R(FN,BIT,R,...) \
	function FN() { \
		this.tmp_byte = R & (0x01 << (BIT)); \
		REG_F = REG_F & FLAG_C | FLAG_H \
				| this.tmp_byte & (FLAG_S | FLAG_5 | FLAG_3) | (this.tmp_byte ? 0 : FLAG_Z) | Cpu.tbl_parity[this.tmp_byte] \
				| R & (FLAG_5 | FLAG_3); \
		__VA_ARGS__ \
	}

#define OP_BIT_MHL(FN,BIT,...) \
	function FN() { \
		DO_BIT_M(GET_HL, BIT); \
		this.tstate += (8-4); \
		__VA_ARGS__ \
	}

#define OP_BIT_PORP(FN,BIT,GET_RP) \
	function FN() { \
		this.tmp_addr = GET_RP + this.tmp_int8 & 0xFFFF; \
		this.SET_MP(this.tmp_addr); \
		DO_BIT_M(this.tmp_addr, BIT); \
		this.tstate += (16-4); \
		DO_PREF_00; \
	}

#define OP_RES_R(FN,BIT,R,...) \
	function FN() { \
		R &= 0xFF ^ (0x01 << (BIT)); \
		__VA_ARGS__ \
	}

#define OP_RES_MHL(FN,BIT,...) \
	function FN() { \
		this.tmp_byte = this.ptr_read(GET_HL); \
		this.tmp_byte &= 0xFF ^ (0x01 << (BIT)); \
		this.ptr_write(GET_HL, this.tmp_byte); \
		this.tstate += (11-4); \
		__VA_ARGS__ \
	}

#define OP_LD_RES_PORP(FN,BIT,R,GET_RP) \
	function FN() { \
		this.tmp_addr = GET_RP + this.tmp_int8 & 0xFFFF; \
		this.SET_MP(this.tmp_addr); \
		this.tmp_byte = this.ptr_read(this.tmp_addr); \
		this.tmp_byte &= 0xFF ^ (0x01 << (BIT)); \
		R = this.tmp_byte; \
		this.ptr_write(this.tmp_addr, this.tmp_byte); \
		this.tstate += (19-4); \
		DO_PREF_00; \
	}

#define OP_RES_PORP(FN,BIT,GET_RP) \
	function FN() { \
		this.tmp_addr = GET_RP + this.tmp_int8 & 0xFFFF; \
		this.SET_MP(this.tmp_addr); \
		this.tmp_byte = this.ptr_read(this.tmp_addr); \
		this.tmp_byte &= 0xFF ^ (0x01 << (BIT)); \
		this.ptr_write(this.tmp_addr, this.tmp_byte); \
		this.tstate += (19-4); \
		DO_PREF_00; \
	}

#define OP_SET_R(FN,BIT,R,...) \
	function FN() { \
		R |= 0x01 << (BIT); \
		__VA_ARGS__ \
	}

#define OP_SET_MHL(FN,BIT,...) \
	function FN() { \
		this.tmp_byte = this.ptr_read(GET_HL); \
		this.tmp_byte |= 0x01 << (BIT); \
		this.ptr_write(GET_HL, this.tmp_byte); \
		this.tstate += (11-4); \
		__VA_ARGS__ \
	}

#define OP_LD_SET_PORP(FN,BIT,R,GET_RP) \
	function FN() { \
		this.tmp_addr = GET_RP + this.tmp_int8 & 0xFFFF; \
		this.SET_MP(this.tmp_addr); \
		this.tmp_byte = this.ptr_read(this.tmp_addr); \
		this.tmp_byte |= 0x01 << (BIT); \
		R = this.tmp_byte; \
		this.ptr_write(this.tmp_addr, this.tmp_byte); \
		this.tstate += (19-4); \
		DO_PREF_00; \
	}

#define OP_SET_PORP(FN,BIT,GET_RP) \
	function FN() { \
		this.tmp_addr = GET_RP + this.tmp_int8 & 0xFFFF; \
		this.SET_MP(this.tmp_addr); \
		this.tmp_byte = this.ptr_read(this.tmp_addr); \
		this.tmp_byte |= 0x01 << (BIT); \
		this.ptr_write(this.tmp_addr, this.tmp_byte); \
		this.tstate += (19-4); \
		DO_PREF_00; \
	}

#define OP_IN_R_BC_P00(FN,R) \
	function FN() { \
		R = this.ptr_in(GET_BC); \
		this.SET_MP(GET_BC + 1 & 0xFFFF); \
		REG_F = REG_F & FLAG_C | R & (FLAG_S | FLAG_5 | FLAG_3) | (R ? 0 : FLAG_Z) | Cpu.tbl_parity[R]; \
		this.tstate += (8-4); \
		DO_PREF_00; \
	}

#define OP_OUT_BC_R_P00(FN,R) \
	function FN() { \
		this.ptr_out(GET_BC, R); \
		this.SET_MP(GET_BC + 1 & 0xFFFF); \
		this.tstate += (8-4); \
		DO_PREF_00; \
	}

#define OP_IN_F_BC_P00(FN) \
	function FN() { \
		this.tmp_byte = this.ptr_in(GET_BC); \
		this.SET_MP(GET_BC + 1 & 0xFFFF); \
		REG_F = REG_F & FLAG_C | this.tmp_byte & (FLAG_S | FLAG_5 | FLAG_3) | (this.tmp_byte ? 0 : FLAG_Z) | Cpu.tbl_parity[this.tmp_byte]; \
		this.tstate += (8-4); \
		DO_PREF_00; \
	}

#define OP_OUT_BC_0_P00(FN) \
	function FN() { \
		this.ptr_out(GET_BC, 0); \
		this.SET_MP(GET_BC + 1 & 0xFFFF); \
		this.tstate += (8-4); \
		DO_PREF_00; \
	}

#define OP_SBC_HL_RP_P00(FN,GET_RP) \
	function FN() { \
		DO_SBC_16(this.SET_HL, GET_HL, GET_RP); \
		this.tstate += (11-4); \
		DO_PREF_00; \
	}

#define OP_NEG_P00(FN) \
	function FN() { \
		this.tmp_byte = REG_A; \
		REG_A = 0; \
		DO_SUB_8(REG_A, this.tmp_byte); \
		DO_PREF_00; \
	}

#define OP_RETN_P00(FN) \
	function FN() { \
		REG_IFF1 = REG_IFF2; \
		DO_RET; \
		this.tstate += (10-4); \
		DO_PREF_00; \
	}

#define OP_RETI_P00 OP_RETN_P00

#define OP_IM_P00(FN,MODE) \
	function FN() { \
		REG_IM = MODE; \
		DO_PREF_00; \
	}

#define OP_LD_I_A_P00(FN) \
	function FN() { \
		REG_I = REG_A; \
		this.tstate += (5-4); \
		DO_PREF_00; \
	}

#define OP_ADC_HL_RP_P00(FN,GET_RP) \
	function FN() { \
		DO_ADC_16(this.SET_HL, GET_HL, GET_RP); \
		this.tstate += (11-4); \
		DO_PREF_00; \
	}

#define OP_LD_RR_A_P00(FN) \
	function FN() { \
		REG_R = REG_A; \
		this.tstate += (5-4); \
		DO_PREF_00; \
	}

#define OP_LD_A_I_P00(FN) \
	function FN() { \
		REG_A = REG_I; \
		REG_F = REG_F & FLAG_C | REG_A & (FLAG_S | FLAG_5 | FLAG_3) | (REG_A ? 0 : FLAG_Z) | REG_IFF2 << FLAG_PV_IFF_S8; \
		this.is_reset_pv = true; \
		this.tstate += (5-4); \
		DO_PREF_00; \
	}

#define OP_LD_A_RR_P00(FN) \
	function FN() { \
		REG_A = REG_R; \
		REG_F = REG_F & FLAG_C | REG_A & (FLAG_S | FLAG_5 | FLAG_3) | (REG_A ? 0 : FLAG_Z) | REG_IFF2 << FLAG_PV_IFF_S8; \
		this.is_reset_pv = true; \
		this.tstate += (5-4); \
		DO_PREF_00; \
	}

#define OP_RRD_P00(FN) \
	function FN() { \
		this.tmp_byte = this.ptr_read(GET_HL); \
		this.ptr_write(GET_HL, REG_A << 4 & 0xFF | this.tmp_byte >> 4); \
		REG_A = REG_A & 0xF0 | this.tmp_byte & 0x0F; \
		REG_F = REG_F & FLAG_C | REG_A & (FLAG_S | FLAG_5 | FLAG_3) | (REG_A ? 0 : FLAG_Z) | Cpu.tbl_parity[REG_A]; \
		this.SET_MP(GET_HL + 1 & 0xFFFF); \
		this.tstate += (14-4); \
		DO_PREF_00; \
	}

#define OP_RLD_P00(FN) \
	function FN() { \
		this.tmp_byte = this.ptr_read(GET_HL); \
		this.ptr_write(GET_HL, this.tmp_byte << 4 & 0xFF | REG_A & 0x0F); \
		REG_A = REG_A & 0xF0 | this.tmp_byte >> 4; \
		REG_F = REG_F & FLAG_C | REG_A & (FLAG_S | FLAG_5 | FLAG_3) | (REG_A ? 0 : FLAG_Z) | Cpu.tbl_parity[REG_A]; \
		this.SET_MP(GET_HL + 1 & 0xFFFF); \
		this.tstate += (14-4); \
		DO_PREF_00; \
	}

#define OP_LDI_P00(FN) \
	function FN() { \
		DO_REP_LD; \
		DO_REP_LD_INC; \
		DO_PREF_00; \
	}

#define OP_LDD_P00(FN) \
	function FN() { \
		DO_REP_LD; \
		DO_REP_LD_DEC; \
		DO_PREF_00; \
	}

#define OP_LDIR_P00(FN) \
	function FN() { \
		DO_REP_LD; \
		DO_REP_LD_INCR; \
		DO_PREF_00; \
	}

#define OP_LDDR_P00(FN) \
	function FN() { \
		DO_REP_LD; \
		DO_REP_LD_DECR; \
		DO_PREF_00; \
	}

#define OP_CPI_P00(FN) \
	function FN() { \
		DO_REP_CP; \
		DO_REP_CP_INC; \
		DO_PREF_00; \
	}

#define OP_CPD_P00(FN) \
	function FN() { \
		DO_REP_CP; \
		DO_REP_CP_DEC; \
		DO_PREF_00; \
	}

#define OP_CPIR_P00(FN) \
	function FN() { \
		DO_REP_CP; \
		DO_REP_CP_INCR; \
		DO_PREF_00; \
	}

#define OP_CPDR_P00(FN) \
	function FN() { \
		DO_REP_CP; \
		DO_REP_CP_DECR; \
		DO_PREF_00; \
	}

#define OP_INI_P00(FN) \
	function FN() { \
		DO_REP_INI; \
		this.tstate += (12-4); \
		DO_PREF_00; \
	}

#define OP_IND_P00(FN) \
	function FN() { \
		DO_REP_IND; \
		this.tstate += (12-4); \
		DO_PREF_00; \
	}

#define OP_INIR_P00(FN) \
	function FN() { \
		DO_REP_INI; \
		DO_REP_IN_OUT_REP; \
		DO_PREF_00; \
	}

#define OP_INDR_P00(FN) \
	function FN() { \
		DO_REP_IND; \
		DO_REP_IN_OUT_REP; \
		DO_PREF_00; \
	}

#define OP_OUTI_P00(FN) \
	function FN() { \
		DO_REP_OUTI; \
		this.tstate += (12-4); \
		DO_PREF_00; \
	}

#define OP_OUTD_P00(FN) \
	function FN() { \
		DO_REP_OUTD; \
		this.tstate += (12-4); \
		DO_PREF_00; \
	}

#define OP_OTIR_P00(FN) \
	function FN() { \
		DO_REP_OUTI; \
		DO_REP_IN_OUT_REP; \
		DO_PREF_00; \
	}

#define OP_OTDR_P00(FN) \
	function FN() { \
		DO_REP_OUTD; \
		DO_REP_IN_OUT_REP; \
		DO_PREF_00; \
	}

#define OP_PREF_CB(FN) \
	function FN() { \
		this.prefix = 0xCB; \
		this.optable = optable_CB; \
		this.is_noint = true; \
	}

#define OP_PREF_DD(FN) \
	function FN() { \
		this.prefix = 0xDD; \
		this.optable = optable_DD; \
		this.is_noint = true; \
	}

#define OP_PREF_ED(FN) \
	function FN() { \
		this.prefix = 0xED; \
		this.optable = optable_ED; \
		this.is_noint = true; \
	}

#define OP_PREF_FD(FN) \
	function FN() { \
		this.prefix = 0xFD; \
		this.optable = optable_FD; \
		this.is_noint = true; \
	}

#define OP_PREF_XX_CB(FN,OPTBL) \
	function FN() { \
		this.tmp_int8 = this.read_offset(); \
		var op = this.read_byte(); \
		OPTBL[op].call(this); \
	}

#define CC_NZ	!(REG_F & FLAG_Z)
#define CC_Z	REG_F & FLAG_Z
#define CC_NC	!(REG_F & FLAG_C)
#define CC_C	REG_F & FLAG_C
#define CC_PO	!(REG_F & FLAG_PV)	// parity odd
#define CC_PE	REG_F & FLAG_PV		// parity even
#define CC_P	!(REG_F & FLAG_S)	// positive
#define CC_M	REG_F & FLAG_S		// minus

#define DO_PREF_00 \
	this.prefix = 0x00; \
	this.optable = optable_00;

#define DO_INC_16(S,G) \
	S(G + 1 & 0xFFFF);

#define DO_DEC_16(S,X) \
	S(X - 1 & 0xFFFF);

#define DO_ADD_16(S,X,Y) \
	this.tmp_dword = X + Y; \
	this.tmp_word_b = (X & 0x0FFF) + (Y & 0x0FFF); \
	this.SET_MP(X + 1 & 0xFFFF); \
	S(this.tmp_dword & 0xFFFF); \
	REG_F = REG_F & (FLAG_PV | FLAG_Z | FLAG_S) \
			| (this.tmp_dword & FLAG_C_M16) >> FLAG_C_S16 \
			| (this.tmp_dword >> 8) & (FLAG_5 | FLAG_3) \
			| (this.tmp_word_b & FLAG_H_M16) >> FLAG_H_S16;

#define DO_ADC_16(S,X,Y) \
	this.tmp_dword = X + Y + (REG_F & FLAG_C); \
	this.tmp_word_b = (X & 0x0FFF) + (Y & 0x0FFF) + (REG_F & FLAG_C); \
	this.tmp_int32 = CONV_INT16(X) + CONV_INT16(Y) + (REG_F & FLAG_C); \
	this.SET_MP(X + 1 & 0xFFFF); \
	S(this.tmp_dword & 0xFFFF); \
	REG_F = (this.tmp_dword & FLAG_C_M16) >> FLAG_C_S16 \
			| (this.tmp_word_b & FLAG_H_M16) >> FLAG_H_S16 \
			| (this.tmp_int32 < -32768 || this.tmp_int32 > 32767 ? FLAG_PV : 0) \
			| ((this.tmp_dword >> 8) & (FLAG_S | FLAG_5 | FLAG_3)) \
			| (this.tmp_dword & 0xFFFF ? 0 : FLAG_Z);

#define DO_SBC_16(S,X,Y) \
	this.tmp_dword = X - Y - (REG_F & FLAG_C) & 0xFFFFFFFF; \
	this.tmp_word_b = (X & 0x0FFF) - (Y & 0x0FFF) - (REG_F & FLAG_C) & 0xFFFF; \
	this.tmp_int32 = CONV_INT16(X) - CONV_INT16(Y) - (REG_F & FLAG_C); \
	REG_F = (this.tmp_dword & FLAG_C_M16) >> FLAG_C_S16 | FLAG_N \
			| (this.tmp_word_b & FLAG_H_M16) >> FLAG_H_S16 \
			| (this.tmp_int32 < -32768 || this.tmp_int32 > 32767 ? FLAG_PV : 0) \
			| (this.tmp_dword >> 8) & (FLAG_S | FLAG_5 | FLAG_3) \
			| (this.tmp_dword & 0xFFFF ? 0 : FLAG_Z); \
	this.SET_MP(X + 1 & 0xFFFF); \
	S(this.tmp_dword & 0xFFFF);

#define DO_INC_8(X) \
	REG_F = REG_F & FLAG_C | ((X & 0x0F) + 1 & FLAG_H); \
	X = X + 1 & 0xFF; \
	REG_F |= (X & FLAG_PV_M8) >> FLAG_PV_S8 | X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z);

#define DO_DEC_8(X) \
	REG_F = REG_F & FLAG_C | FLAG_N | ((X & 0x0F) - 1 & FLAG_H) | (X & FLAG_PV_M8) >> FLAG_PV_S8; \
	X = X - 1 & 0xFF; \
	REG_F |= X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z);

#define DO_ADD_8(X,Y) \
	this.tmp_word = X + Y; \
	this.tmp_byte_b = (X & 0x0F) + (Y & 0x0F); \
	this.tmp_int16 = CONV_INT8(X) + CONV_INT8(Y); \
	REG_F = (this.tmp_word & FLAG_C_M8) >> FLAG_C_S8 \
			| this.tmp_byte_b & FLAG_H \
			| (this.tmp_int16 < -128 || this.tmp_int16 > 127 ? FLAG_PV : 0) \
			| this.tmp_word & (FLAG_S | FLAG_5 | FLAG_3) \
			| (this.tmp_word & 0xFF ? 0 : FLAG_Z); \
	X = this.tmp_word & 0xFF;

#define DO_ADC_8(X,Y) \
	this.tmp_word = X + Y + (REG_F & FLAG_C); \
	this.tmp_byte_b = (X & 0x0F) + (Y & 0x0F) + (REG_F & FLAG_C); \
	this.tmp_int16 = CONV_INT8(X) + CONV_INT8(Y) + (REG_F & FLAG_C); \
	REG_F = (this.tmp_word & FLAG_C_M8) >> FLAG_C_S8 \
			| this.tmp_byte_b & FLAG_H \
			| (this.tmp_int16 < -128 || this.tmp_int16 > 127 ? FLAG_PV : 0) \
			| this.tmp_word & (FLAG_S | FLAG_5 | FLAG_3) \
			| (this.tmp_word & 0xFF ? 0 : FLAG_Z); \
	X = this.tmp_word & 0xFF;

#define DO_SUB_8(X,Y) \
	this.tmp_word = X - Y & 0xFFFF; \
	this.tmp_byte_b = (X & 0x0F) - (Y & 0x0F) & 0xFF; \
	this.tmp_int16 = CONV_INT8(X) - CONV_INT8(Y); \
	REG_F = (this.tmp_word & FLAG_C_M8) >> FLAG_C_S8 | FLAG_N \
			| this.tmp_byte_b & FLAG_H \
			| (this.tmp_int16 < -128 || this.tmp_int16 > 127 ? FLAG_PV : 0) \
			| this.tmp_word & (FLAG_S | FLAG_5 | FLAG_3) \
			| (this.tmp_word & 0xFF ? 0 : FLAG_Z); \
	X = this.tmp_word & 0xFF;

#define DO_SBC_8(X,Y) \
	this.tmp_word = X - Y - (REG_F & FLAG_C) & 0xFFFF; \
	this.tmp_byte_b = (X & 0x0F) - (Y & 0x0F) - (REG_F & FLAG_C) & 0xFF; \
	this.tmp_int16 = CONV_INT8(X) - CONV_INT8(Y) - (REG_F & FLAG_C); \
	REG_F = (this.tmp_word & FLAG_C_M8) >> FLAG_C_S8 | FLAG_N \
			| this.tmp_byte_b & FLAG_H \
			| (this.tmp_int16 < -128 || this.tmp_int16 > 127 ? FLAG_PV : 0) \
			| this.tmp_word & (FLAG_S | FLAG_5 | FLAG_3) \
			| (this.tmp_word & 0xFF ? 0 : FLAG_Z); \
	X = this.tmp_word & 0xFF;

#define DO_AND_8(X,Y) \
	X &= Y; \
	REG_F = FLAG_H | (X & (FLAG_S | FLAG_5 | FLAG_3)) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_XOR_8(X,Y) \
	X ^= Y; \
	REG_F = X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_OR_8(X,Y) \
	X |= Y; \
	REG_F = X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_CP_8(X,Y) \
	this.tmp_word = X - Y & 0xFFFF; \
	this.tmp_byte_b = (X & 0x0F) - (Y & 0x0F) & 0xFF; \
	this.tmp_int16 = CONV_INT8(X) - CONV_INT8(Y); \
	REG_F = (this.tmp_word & FLAG_C_M8) >> FLAG_C_S8 | FLAG_N \
			| this.tmp_byte_b & FLAG_H \
			| (this.tmp_int16 < -128 || this.tmp_int16 > 127 ? FLAG_PV : 0) \
			| Y & (FLAG_5 | FLAG_3) \
			| this.tmp_word & FLAG_S \
			| (this.tmp_word & 0xFF ? 0 : FLAG_Z);

#define DO_POP_TMP \
	this.tmp_word = this.ptr_read(GET_SP); \
	this.SET_SP(GET_SP + 1 & 0xFFFF); \
	this.tmp_word |= this.ptr_read(GET_SP) << 8; \
	this.SET_SP(GET_SP + 1 & 0xFFFF);

#define DO_RET \
	DO_POP_TMP; \
	this.SET_PC(this.tmp_word); \
	this.SET_MP(this.tmp_word);

#define DO_PUSH(VAL) \
	this.tmp_word_b = VAL; \
	this.SET_SP(GET_SP - 1 & 0xFFFF); this.ptr_write(GET_SP, this.tmp_word_b >> 8); \
	this.SET_SP(GET_SP - 1 & 0xFFFF); this.ptr_write(GET_SP, this.tmp_word_b & 0xFF);

#define DO_OUT(PORT,VAL) \
	this.ptr_out(PORT, VAL); \
	REG_MPL = PORT + 1 & 0xFF; \
	REG_MPH = REG_A;

#define DO_IN(VAL,PORT) \
	VAL = this.ptr_in(PORT); \
	this.SET_MP(PORT + 1 & 0xFFFF);

#define DO_RLC_8(X) \
	X = X << 1 & 0xFF | X >> 7; \
	REG_F = X & (FLAG_S | FLAG_5 | FLAG_3 | FLAG_C) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_RRC_8(X) \
	REG_F = X & FLAG_C; \
	X = X >> 1 | X << 7 & 0xFF; \
	REG_F |= X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_RL_8(X) \
	this.tmp_byte_b = X; \
	X = X << 1 & 0xFF | REG_F & FLAG_C; \
	REG_F = this.tmp_byte_b >> 7 | X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_RR_8(X) \
	this.tmp_byte_b = X; \
	X = X >> 1 | REG_F << 7 & 0xFF; \
	REG_F = this.tmp_byte_b & FLAG_C | X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_SLA_8(X) \
	REG_F = X >> 7; \
	X = X << 1 & 0xFF; \
	REG_F |= X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_SRA_8(X) \
	REG_F = X & FLAG_C; \
	X = X & 0x80 | X >> 1; \
	REG_F |= X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_SLL_8(X) \
	REG_F = X >> 7; \
	X = X << 1 & 0xFF | 0x01; \
	REG_F |= X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_SRL_8(X) \
	REG_F = X & FLAG_C; \
	X >>= 1; \
	REG_F |= X & (FLAG_S | FLAG_5 | FLAG_3) | (X ? 0 : FLAG_Z) | Cpu.tbl_parity[X];

#define DO_BIT_M(X,BIT) \
	this.tmp_byte = this.ptr_read(X); \
	this.tmp_byte_b = this.tmp_byte & (0x01 << (BIT)); \
	REG_F = REG_F & FLAG_C | FLAG_H \
			| this.tmp_byte_b & FLAG_S | (this.tmp_byte_b ? 0 : FLAG_Z) | Cpu.tbl_parity[this.tmp_byte_b] \
			| REG_MPH & (FLAG_5 | FLAG_3);

#define DO_REP_LD \
	this.tmp_byte = this.ptr_read(GET_HL); \
	this.ptr_write(GET_DE, this.tmp_byte); \
	this.SET_BC(GET_BC - 1 & 0xFFFF); \
	this.tmp_byte += REG_A; \
	REG_F = REG_F & (FLAG_S | FLAG_Z | FLAG_C) | (GET_BC ? FLAG_PV : 0) \
			| this.tmp_byte & FLAG_3 | (this.tmp_byte << FLAG_N_TO_5) & FLAG_5;

#define DO_REP_LD_INC \
	this.SET_HL(GET_HL + 1 & 0xFFFF); \
	this.SET_DE(GET_DE + 1 & 0xFFFF); \
	this.tstate += (12-4);

#define DO_REP_LD_DEC \
	this.SET_HL(GET_HL - 1 & 0xFFFF); \
	this.SET_DE(GET_DE - 1 & 0xFFFF); \
	this.tstate += (12-4);

#define DO_REP_LD_INCR \
	this.SET_HL(GET_HL + 1 & 0xFFFF); \
	this.SET_DE(GET_DE + 1 & 0xFFFF); \
	if (GET_BC) { \
		this.SET_PC(GET_PC - 2 & 0xFFFF); \
		this.SET_MP(GET_PC + 1 & 0xFFFF); \
		this.tstate += (17-4); \
	} else { \
		this.tstate += (12-4); \
	}

#define DO_REP_LD_DECR \
	this.SET_HL(GET_HL - 1 & 0xFFFF); \
	this.SET_DE(GET_DE - 1 & 0xFFFF); \
	if (GET_BC) { \
		this.SET_PC(GET_PC - 2 & 0xFFFF); \
		this.SET_MP(GET_PC + 1 & 0xFFFF); \
		this.tstate += (17-4); \
	} else { \
		this.tstate += (12-4); \
	}

#define DO_REP_CP \
	this.tmp_byte = this.ptr_read(GET_HL); \
	this.tmp_byte_b = (REG_A & 0x0F) - (this.tmp_byte & 0x0F) & 0xFF; \
	this.tmp_byte = REG_A - this.tmp_byte & 0xFF; \
	this.SET_BC(GET_BC - 1 & 0xFFFF); \
	REG_F = REG_F & FLAG_C | FLAG_N | (GET_BC ? FLAG_PV : 0) \
			| this.tmp_byte_b & FLAG_H | (this.tmp_byte ? 0 : FLAG_Z) | this.tmp_byte & FLAG_S; \
	this.tmp_byte = this.tmp_byte - ((REG_F & FLAG_H) >> FLAG_H_TO_C) & 0xFF; \
	REG_F |= this.tmp_byte & FLAG_3 | (this.tmp_byte << FLAG_N_TO_5) & FLAG_5;

#define DO_REP_CP_INC \
	this.SET_HL(GET_HL + 1 & 0xFFFF); \
	this.SET_MP(GET_MP + 1 & 0xFFFF); \
	this.tstate += (12-4);

#define DO_REP_CP_DEC \
	this.SET_HL(GET_HL - 1 & 0xFFFF); \
	this.SET_MP(GET_MP - 1 & 0xFFFF); \
	this.tstate += (12-4);

#define DO_REP_CP_INCR \
	this.SET_HL(GET_HL + 1 & 0xFFFF); \
	if ((REG_F & (FLAG_Z | FLAG_PV)) == FLAG_PV) { \
		this.SET_PC(GET_PC - 2 & 0xFFFF); \
		this.SET_MP(GET_PC + 1 & 0xFFFF); \
		this.tstate += (17-4); \
	} else { \
		this.SET_MP(GET_MP + 1 & 0xFFFF); \
		this.tstate += (12-4); \
	}

#define DO_REP_CP_DECR \
	this.SET_HL(GET_HL - 1 & 0xFFFF); \
	if ((REG_F & (FLAG_Z | FLAG_PV)) == FLAG_PV) { \
		this.SET_PC(GET_PC - 2 & 0xFFFF); \
		this.SET_MP(GET_PC + 1 & 0xFFFF); \
		this.tstate += (17-4); \
	} else { \
		this.SET_MP(GET_MP + 1 & 0xFFFF); \
		this.tstate += (12-4); \
	}

#define DO_REP_INI \
	this.tmp_byte = this.ptr_in(GET_BC); \
	this.ptr_write(GET_HL, this.tmp_byte); \
	this.SET_MP(GET_BC + 1 & 0xFFFF); \
	REG_B = REG_B - 1 & 0xFF; \
	this.SET_HL(GET_HL + 1 & 0xFFFF); \
	this.tmp_word = this.tmp_byte + (REG_C + 1 & 0xFF); \
	REG_F = (this.tmp_byte >> FLAG_S_TO_N) & FLAG_N \
			| REG_B & (FLAG_S | FLAG_5 | FLAG_3) \
			| (REG_B ? 0 : FLAG_Z) \
			| Cpu.tbl_parity[(this.tmp_word & 0x07) ^ REG_B] \
			| (this.tmp_word > 255 ? (FLAG_C | FLAG_H) : 0);

#define DO_REP_IND \
	this.tmp_byte = this.ptr_in(GET_BC); \
	this.ptr_write(GET_HL, this.tmp_byte); \
	this.SET_MP(GET_BC - 1 & 0xFFFF); \
	REG_B = REG_B - 1 & 0xFF; \
	this.SET_HL(GET_HL - 1 & 0xFFFF); \
	this.tmp_word = this.tmp_byte + (REG_C - 1 & 0xFF); \
	REG_F = (this.tmp_byte >> FLAG_S_TO_N) & FLAG_N \
			| REG_B & (FLAG_S | FLAG_5 | FLAG_3) \
			| (REG_B ? 0 : FLAG_Z) \
			| Cpu.tbl_parity[(this.tmp_word & 0x07) ^ REG_B] \
			| (this.tmp_word > 255 ? (FLAG_C | FLAG_H) : 0);

#define DO_REP_IN_OUT_REP \
	if (REG_B) { \
		this.SET_PC(GET_PC - 2 & 0xFFFF); \
		this.tstate += (17-4); \
	} else { \
		this.tstate += (12-4); \
	}

#define DO_REP_OUTI \
	this.tmp_byte = this.ptr_read(GET_HL); \
	REG_B = REG_B - 1 & 0xFF; \
	this.SET_MP(GET_BC + 1 & 0xFFFF); \
	this.ptr_out(GET_BC, this.tmp_byte); \
	this.SET_HL(GET_HL + 1 & 0xFFFF); \
	this.tmp_word = this.tmp_byte + REG_L; \
	REG_F = (this.tmp_byte >> FLAG_S_TO_N) & FLAG_N \
			| REG_B & (FLAG_S | FLAG_5 | FLAG_3) \
			| (REG_B ? 0 : FLAG_Z) \
			| Cpu.tbl_parity[(this.tmp_word & 0x07) ^ REG_B] \
			| (this.tmp_word > 255 ? (FLAG_C | FLAG_H) : 0);

#define DO_REP_OUTD \
	this.tmp_byte = this.ptr_read(GET_HL); \
	REG_B = REG_B - 1 & 0xFF; \
	this.SET_MP(GET_BC - 1 & 0xFFFF); \
	this.ptr_out(GET_BC, this.tmp_byte); \
	this.SET_HL(GET_HL - 1 & 0xFFFF); \
	this.tmp_word = this.tmp_byte + REG_L; \
	REG_F = (this.tmp_byte >> FLAG_S_TO_N) & FLAG_N \
			| REG_B & (FLAG_S | FLAG_5 | FLAG_3) \
			| (REG_B ? 0 : FLAG_Z) \
			| Cpu.tbl_parity[(this.tmp_word & 0x07) ^ REG_B] \
			| (this.tmp_word > 255 ? (FLAG_C | FLAG_H) : 0);
